/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) Under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for Additional information regarding copyright ownership.
   The ASF licenses this file to You Under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed Under the License is distributed on an "AS Is" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations Under the License.
==================================================================== */

namespace NPOI.SS.Formula.PTG
{
    using System;
    
    using NPOI.Util;
    using NPOI.HSSF.UserModel;
    using NPOI.SS.UserModel;

    /**
     * @author Daniel Noll (daniel at nuix dot com dot au)
     */
    public class ErrPtg : ScalarConstantPtg
    {

        // convenient access to namespace
        //private static HSSFErrorConstants EC = null;

        /** <b>#NULL!</b>  - Intersection of two cell ranges is empty */
        public static readonly ErrPtg NULL_INTERSECTION = new ErrPtg(FormulaError.NULL.Code);
        /** <b>#DIV/0!</b> - Division by zero */
        public static readonly ErrPtg DIV_ZERO = new ErrPtg(FormulaError.DIV0.Code);
        /** <b>#VALUE!</b> - Wrong type of operand */
        public static readonly ErrPtg VALUE_INVALID = new ErrPtg(FormulaError.VALUE.Code);
        /** <b>#REF!</b> - Illegal or deleted cell reference */
        public static readonly ErrPtg REF_INVALID = new ErrPtg(FormulaError.REF.Code);
        /** <b>#NAME?</b> - Wrong function or range name */
        public static readonly ErrPtg NAME_INVALID = new ErrPtg(FormulaError.NAME.Code);
        /** <b>#NUM!</b> - Value range overflow */
        public static readonly ErrPtg NUM_ERROR = new ErrPtg(FormulaError.NUM.Code);
        /** <b>#N/A</b> - Argument or function not available */
        public static readonly ErrPtg N_A = new ErrPtg(FormulaError.NA.Code);

        public static ErrPtg ValueOf(int code)
        {
            switch ((FormulaErrorEnum)code)
            {
                case FormulaErrorEnum.DIV_0: return DIV_ZERO;
                case FormulaErrorEnum.NA: return N_A;
                case FormulaErrorEnum.NAME: return NAME_INVALID;
                case FormulaErrorEnum.NULL: return NULL_INTERSECTION;
                case FormulaErrorEnum.NUM: return NUM_ERROR;
                case FormulaErrorEnum.REF: return REF_INVALID;
                case FormulaErrorEnum.VALUE: return VALUE_INVALID;
                default:
                    throw new InvalidOperationException("Unexpected error code (" + code + ")");
            }
        }

        public const byte sid = 0x1c;
        private const int SIZE = 2;
        private int field_1_error_code;

        /** Creates new ErrPtg */

        public ErrPtg(int errorCode)
        {
            if (!FormulaError.IsValidCode(errorCode))
            {
                throw new ArgumentException("Invalid error code (" + errorCode + ")");
            }
            field_1_error_code = errorCode;
        }

        public ErrPtg(ILittleEndianInput in1)
            : this(in1.ReadByte())
        {
            
        }

        public override void Write(ILittleEndianOutput out1)
        {
            out1.WriteByte(sid + PtgClass);
            out1.WriteByte((byte)field_1_error_code);
        }

        public override String ToFormulaString()
        {
            return FormulaError.ForInt(field_1_error_code).String;
        }

        public override int Size
        {
            get { return SIZE; }
        }

        public int ErrorCode
        {
            get { return field_1_error_code; }
        }
    }
}